/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/


/************************************************************
 *
 * vm_assert.h --
 *
 *  see http://vmweb.vmware.com/~mts/WebSite/guide/programming/asserts.html for
 *  proper usage
 *
 ************************************************************/


#ifndef VM_ASSERT_H
#define VM_ASSERT_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MONITOR
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

#include "vm_basic_types.h"

/*
 * Compact filecodes for the monitor.
 */
#define VMSTRINGIFY(x) #x
#define INT_TO_FILECODE(x) VMSTRINGIFY(F(x))
#ifdef FILECODEINT
#define FILECODE INT_TO_FILECODE(FILECODEINT)
#endif

#ifndef FILECODEINT
#ifdef MONITOR
#error Monitor must use FILECODEINT
#else
#define FILECODEINT 0
#endif
#endif


#if defined(VMX86_DEVEL)
#ifdef FILE_BASENAME
#define FILECODESTRING  FILE_BASENAME
#else
#define FILECODESTRING  __FILE__
#endif
#else
#ifndef FILECODE
#error <FILECODE> not defined.  Update ".../support/scripts/filecodename.prl".
#endif
#define FILECODESTRING FILECODE 
#endif

EXTERN NORETURN_DECL(void Panic(char *, ...)) PRINTF_DECL(1, 2);
EXTERN NORETURN_DECL(void PanicLocation(const char *,
                                        char *,...)) PRINTF_DECL(2, 3);
EXTERN NORETURN_DECL(void Panic_CoreDump(void));

EXTERN void Log_SetPanicMsgPost(Bool postMsg);
EXTERN Bool Log_GetPanicMsgPost(void);
EXTERN void Warning(char *, ...) PRINTF_DECL(1, 2);
EXTERN void Log(char *, ...) PRINTF_DECL(1, 2);
EXTERN void WarningThrottled(uint32 *count, char *, ...) PRINTF_DECL(2, 3);
EXTERN void LogThrottled(uint32 *count, char *, ...) PRINTF_DECL(2, 3);
/* DB family:  messages which are parsed by logfile database system */
#define WarningDB Warning
#define LogDB Log
#define WarningThrottledDB WarningThrottled
#define LogThrottledDB LogThrottled



#define AssertPanicStr             "PANIC %s:%d\n"
#define AssertNotImplStr           "NOT_IMPLEMENTED %s:%d\n"
#define AssertBugStr               "BUG %s:%d bugNr=%d\n"
#define AssertNotImplBugStr        "NOT IMPLEMENTED %s:%d bugNr=%d\n"
#define AssertPanicBugStr          "PANIC %s:%d bugNr=%d\n"
#define AssertNotReachedBugStr     "NOT_REACHED %s:%d bugNr=%d\n"
#define AssertAssertFailedStr      "ASSERT FAILED %s:%d\n"
#define AssertNotReachedFailedStr  "NOTREACHED FAILED %s:%d\n"
#define AssertAssertFailureStr     "ASSERT failure %s:%d r=0x%x e=0x%x\n"
#define AssertUnexpectedWarningStr "UNEXPECTED %d (warning) %s:%d\n"
#define AssertNotTestedWarningStr  "NOT TESTED (warning) %s:%d\n"
#define AssertVMX86AssertFailedStr "VMX86_DEVEL ASSERT FAILED %s:%d\n"

#ifdef MONITOR
extern char Assert_Panic[];
extern char Assert_NotImpl[];
extern char Assert_Bug[];
extern char Assert_NotImplBug[];
extern char Assert_PanicBug[];
extern char Assert_NotReachedBug[];
extern char Assert_AssertFailed[];
extern char Assert_NotReachedFailed[];
extern char Assert_AssertFailure[];
extern char Assert_UnexpectedWarning[];
extern char Assert_NotTestedWarning[];
extern char Assert_VMX86AssertFailed[];
#else /* MONITOR */
#define Assert_Panic AssertPanicStr
#define Assert_NotImpl AssertNotImplStr
#define Assert_Bug AssertBugStr
#define Assert_NotImplBug AssertNotImplBugStr
#define Assert_PanicBug AssertPanicBugStr
#define Assert_NotReachedBug AssertNotReachedBugStr
#define Assert_AssertFailed AssertAssertFailedStr
#define Assert_NotReachedFailed AssertNotReachedFailedStr
#define Assert_AssertFailure AssertAssertFailureStr
#define Assert_UnexpectedWarning AssertUnexpectedWarningStr
#define Assert_NotTestedWarning AssertNotTestedWarningStr
#define Assert_VMX86AssertFailed AssertVMX86AssertFailedStr
#endif

/*
 * Compilation options
 */

#if (defined(VMX86_BETA) || defined(VMX86_RELEASE)) && defined(MONITOR)
#define VMX86_DENSE_ASSERTS
EXTERN NORETURN_DECL(void PanicNotReached(const char *location));
EXTERN NORETURN_DECL(void PanicNotImplemented(const char *location));
EXTERN NORETURN_DECL(void PanicAssert(const char *location));
#endif


#if defined(VMX86_DEBUG) || defined(USERLEVEL)
#define _VMX86_HAS_ASSERTS
#endif


/*
 * Internal macros
 */

#ifdef VMX86_DENSE_ASSERTS
#define STRINGIFY2(x) VMSTRINGIFY(x)              // gotta love cpp
#define LINE_AS_STRING() STRINGIFY2(__LINE__)

#   define _ASSERT_PANIC(_c, _s, _f) do {                                    \
   if (UNLIKELY(! (_c))) {                                                   \
      static char _location_[]                                               \
           __attribute__((section (".locations"))) =                         \
           __FILE__ ":" LINE_AS_STRING();                                    \
      (_f)(_location_);                                                      \
   }                                                                         \
} while (0)
#else
#   define _ASSERT_PANIC(_c, _s, _f) do {    \
   if (UNLIKELY(! (_c))) {                   \
      Panic((_s), FILECODESTRING, __LINE__); \
   }                                         \
} while (0)
#endif

#define _ASSERT_WARNING(_c, _s) do {           \
   if (UNLIKELY(! (_c))) {                     \
      Warning((_s), FILECODESTRING, __LINE__); \
   }                                           \
} while (0)

#define _ASSERT_LOG(_c, _s) do {           \
   if (UNLIKELY(! (_c))) {                 \
      Log((_s), FILECODESTRING, __LINE__); \
   }                                       \
} while (0)

#define _ASSERT_WARNING_BUG(_c, _s, _b) do {         \
   if (UNLIKELY(! (_c))) {                           \
      Warning((_s), (_b), FILECODESTRING, __LINE__); \
   }                                                 \
} while (0)

#define _ASSERT_LOG_BUG(_c, _s, _b) do {         \
   if (UNLIKELY(! (_c))) {                       \
      Log((_s), (_b), FILECODESTRING, __LINE__); \
   }                                             \
} while (0)

#define PANIC() Panic(Assert_Panic,FILECODESTRING, __LINE__)


#define ASSERT_NOT_IMPLEMENTED(_c) _ASSERT_PANIC(_c,Assert_NotImpl,PanicNotImplemented)
#define NOT_IMPLEMENTED()          _ASSERT_PANIC(0,Assert_NotImpl,PanicNotImplemented)

#define ASSERT_BUG(_bugNr,_c) if (UNLIKELY(!(_c))) {              \
   Panic(Assert_Bug, FILECODESTRING, __LINE__,_bugNr);\
}

#define ASSERT_NOT_IMPLEMENTED_BUG(_bugNr,_c) if (UNLIKELY(!(_c))) {          \
   Panic(Assert_NotImplBug, FILECODESTRING, __LINE__,_bugNr);\
}

#define NOT_IMPLEMENTED_BUG(_bugNr)  {                            \
    Panic(Assert_NotImplBug, FILECODESTRING, __LINE__,_bugNr);\
}

#define PANIC_BUG(_bugNr)  {                            \
    Panic(Assert_PanicBug, FILECODESTRING, __LINE__,_bugNr);\
}

#define NOT_REACHED_BUG(_bugNr) {                                          \
    Panic(Assert_NotReachedBug, FILECODESTRING, __LINE__,_bugNr);\
}   
   
#define ASSERT_NOT_IMPLEMENTED_PERROR(_cond) {                    \
   if (UNLIKELY(!(_cond))) {                                      \
      Panic("Unimplemented system call return value %s:%d (%s)\n",\
            FILECODESTRING,__LINE__, Msg_ErrString());            \
   }                                                              \
}


/*
 * Following asserts are present only in the debug versions
 */

#define NOT_TESTED_WARNING()  NOT_TESTED()

#ifdef VMX86_DEBUG
#undef ASSERT
#define ASSERT(_c)     _ASSERT_PANIC(_c,Assert_AssertFailed,PanicAssert)

#define NOT_REACHED()  _ASSERT_PANIC(0,Assert_NotReachedFailed,PanicNotReached)

#define ASSERT_BUG_DEBUGONLY(_bug,_c) ASSERT_BUG(_bug,_c)

#define ASSERT_MESSAGE(_c,_msg) {                                             \
   if (UNLIKELY(!(_c))) {                                                     \
      Warning _msg;                                                           \
      Panic(Assert_AssertFailed, FILECODESTRING, __LINE__);                   \
   }                                                                          \
}

#ifdef VMX86_DEVEL
#define ASSERT_LENGTH(_real,_expected) {                                      \
   if (UNLIKELY((_real) != (_expected))) {                                    \
      Panic(Assert_AssertFailure,                                             \
            FILECODESTRING, __LINE__, _real,_expected);                       \
   }                                                                          \
}
#else  /* !VMX86_DEVEL */
#define ASSERT_LENGTH(_real,_expected) ASSERT((_real) == (_expected))
#endif /* !VMX86_DEVEL */

#else /* VMX86_DEBUG */

#ifndef WINNT_DDK
#undef ASSERT
#define ASSERT(_c)
#endif

#define NOT_REACHED()
#define ASSERT_MESSAGE(_c,_msg)
#define ASSERT_LENGTH(_real,_expected)
#define ASSERT_BUG_DEBUGONLY(_bug,_c)
#endif /* !VMX86_DEBUG */


/*
 * LOG_UNEXPECTED(bugNr) 
 */
#ifdef VMX86_DEBUG
#if defined(VMX86_BETA)
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c) \
   _ASSERT_LOG_BUG(_c, Assert_UnexpectedWarning, _bugNr)
#define LOG_UNEXPECTED(_bugNr) \
   _ASSERT_LOG_BUG(0,  Assert_UnexpectedWarning, _bugNr)
#elif defined(VMX86_RELEASE)
#error
#else
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c) \
   _ASSERT_WARNING_BUG(_c, Assert_UnexpectedWarning, _bugNr)
#define LOG_UNEXPECTED(_bugNr) \
   _ASSERT_WARNING_BUG(0,  Assert_UnexpectedWarning, _bugNr)
#endif
#else  /* VMX86_DEBUG */
#define ASSERT_LOG_UNEXPECTED(_bugNr, _c)
#define LOG_UNEXPECTED(_bugNr)
#endif /* VMX86_DEBUG */

/*
 * Different NOT_TESTED() for VMX86_BETA 
 */
#ifdef VMX86_DEBUG
#if defined(VMX86_BETA)
#define ASSERT_NOT_TESTED(_c) _ASSERT_LOG(_c,Assert_NotTestedWarning)
#define NOT_TESTED()          _ASSERT_LOG(0, Assert_NotTestedWarning)
#elif defined(VMX86_RELEASE)
#error
#else
#define ASSERT_NOT_TESTED(_c) _ASSERT_WARNING(_c,Assert_NotTestedWarning)
#define NOT_TESTED()          _ASSERT_WARNING(0, Assert_NotTestedWarning)
#endif

#define NOT_TESTED_ONCE()                                               \
   do {                                                                 \
      static Bool alreadyPrinted = FALSE;                               \
      if (UNLIKELY(!alreadyPrinted)) { alreadyPrinted = TRUE; NOT_TESTED(); } \
   } while (0)

#define NOT_TESTED_1024()                                               \
   do {                                                                 \
      static uint16 count = 0;                                          \
      if (UNLIKELY(count == 0)) { NOT_TESTED(); }                       \
      count = (count + 1) & 1023;                                       \
   } while (0)

#else  /* VMX86_DEBUG */
#define ASSERT_NOT_TESTED(_c)
#define NOT_TESTED()
#define NOT_TESTED_ONCE()
#define NOT_TESTED_1024()

#endif /* VMX86_DEBUG */


#define ASSERT_NO_INTERRUPTS()   ASSERT(!INTERRUPTS_ENABLED())
#define ASSERT_HAS_INTERRUPTS()  ASSERT(INTERRUPTS_ENABLED())

/*
 * The following are only present for the OBJDIR=obj build (development).
 */
#if defined(VMX86_DEVEL) && defined(VMX86_DEBUG)
#define ASSERT_DEVEL(_c)      _ASSERT_PANIC(_c,Assert_VMX86AssertFailed, \
                                            PanicAssert)

#define MEMORY_LEAK(_ptr) {                  \
   Log("Memory leak 0x%8X at %s:%d\n",       \
       _ptr,FILECODESTRING,__LINE__);        \
}

#else /* VMX86_DEBUG  && VMX86_DEVEL */

#define ASSERT_DEVEL(_c)
#define MEMORY_LEAK()
#endif /* !VMX86_DEBUG || !VMX86_DEVEL */

/* Convenience macros and definitions. Can often be used instead of #ifdef. */
#ifdef VMX86_DEBUG
#undef DEBUG_ONLY
#define vmx86_debug   1
#define DEBUG_ONLY(x) x
#else
#undef DEBUG_ONLY
#define vmx86_debug   0
#define DEBUG_ONLY(x)
#endif

#ifdef VMX86_STATS
#define vmx86_stats   1
#define STATS_ONLY(x) x
#else
#define vmx86_stats   0
#define STATS_ONLY(x)
#endif

#ifdef VMX86_DEVEL
#define vmx86_devel   1
#define DEVEL_ONLY(x) x
#else
#define vmx86_devel   0
#define DEVEL_ONLY(x)
#endif

#ifdef VMX86_LOG
#define vmx86_log     1
#define LOG_ONLY(x)   x
#else
#define vmx86_log     0
#define LOG_ONLY(x)
#endif

#ifdef VMX86_SERVER
#define vmx86_server 1
#define SERVER_ONLY(x) x
#define HOSTED_ONLY(x)
#else
#define vmx86_server 0
#define SERVER_ONLY(x)
#define HOSTED_ONLY(x) x
#endif

#ifdef VMKERNEL
#define vmkernel 1
#define VMKERNEL_ONLY(x) x
#else 
#define vmkernel 0
#define VMKERNEL_ONLY(x)
#endif

#ifdef WIN32
#define WIN32_ONLY(x) x
#define POSIX_ONLY(x)
#else
#define WIN32_ONLY(x)
#define POSIX_ONLY(x) x
#endif

#ifdef MONITOR
#define MONITOR_ONLY(x) x
#define USER_ONLY(x)
#else
#define MONITOR_ONLY(x)
#define USER_ONLY(x) x
#endif
#endif /* VM_ASSERT_H */
