/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#   define FILECODE "F(300)"

/* Must come before any kernel header file --hpreg */
#include "driver-config.h"

#ifdef KERNEL_2_1
#define EXPORT_SYMTAB
#endif

extern int errno;
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/file.h>
#ifdef KERNEL_2_2
#   include <linux/slab.h>
#else
#   include <linux/malloc.h>
#endif
#include <linux/interrupt.h>
#include <linux/wrapper.h>

#ifdef __SMP__
#include <linux/smp.h>
#include <linux/smp_lock.h>
#endif


#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 1, 0)
#   include <linux/poll.h>
#endif

#ifdef DO_APM
#include <linux/apm_bios.h>
#endif

#define __KERNEL_SYSCALLS__
#include <asm/unistd.h>
#include <asm/io.h>

#include "driver.h"
#include "machine.h"
#include "vm_types.h"
#include "vm_assert.h"
#include "modulecall.h"
#include "vm_asm.h"
#include "vmx86.h"
#include "initblock.h"
#include "task.h"
#include "speaker_reg.h"
#include "vtrace.h"
#include "memtrack.h"
#include "task.h"
#include "cpuid.h"

#ifdef VMX86_DEVEL
#include "private.h"
#endif

#ifdef USE_PERFCTRS_HOSTED
#include "perfctr.h"
#endif

#include "hostif.h"
#include "vmhost.h"

#include "compat_file.h"
#include "compat_wait.h"
#include "compat_uaccess.h"
#include "compat_page.h"
#include "compat_wrapper.h"
#include "compat_timer.h"
#include "compat_mm.h"
#include "vmmonInt.h"


int errno;       // required for _exit()
static void LinuxDriverQueue(VMLinux *vmLinux);
static void LinuxDriverDequeue(VMLinux *vmLinux);
static Bool LinuxDriverCheckPadding(void);

#ifdef DO_APM
static int LinuxDriverAPMCallback(apm_event_t event);
static int LinuxDriverAPMstate = APM_STATE_READY;
#endif

#ifndef SUPPORT_LINUXVMWARE
#define VMWare_SetVTracer(VTrace_Set)
#endif

struct VMXLinuxState linuxState;


/*
 *----------------------------------------------------------------------
 *
 * Device Driver Interface --
 *
 *      Runs the VM by implementing open/close/ioctl functions
 *
 *
 *----------------------------------------------------------------------
 */
static int LinuxDriver_Open(struct inode * inode, struct file * filp);

static int LinuxDriver_Ioctl( struct inode *inode, struct file *filp,
         u_int iocmd, unsigned long ioarg );


#ifdef KERNEL_2_1
static int LinuxDriver_Close( struct inode *inode, struct file *filp );
static unsigned int LinuxDriverPoll(struct file *file, poll_table *wait);
#else
static void LinuxDriver_Close( struct inode *inode, struct file *filp );
#endif
#ifdef KERNEL_2_4_0
static struct page *LinuxDriverNoPage(struct vm_area_struct *vma,
			   unsigned long address, int unused);
#else
static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,
			   unsigned long address, int write_access);
#endif
static int LinuxDriverMmap(struct file *filp, struct vm_area_struct *vma);

static void LinuxDriverPollTimeout(unsigned long clientData);

static struct vm_operations_struct vmuser_mops = {
	.nopage = LinuxDriverNoPage
};

static struct file_operations vmuser_fops;


/*
 *----------------------------------------------------------------------
 *
 * VMX86_RegisterMonitor --
 *
 *      (debugging support) Should be the first function of this file
 *
 * Results:
 *
 *      Registers the module.
 *      /sbin/ksyms -a | grep VMX86_RegisterMonitor will return the base
 *      address of that function as loaded in the kernel.
 *
 *      Since this is the first function of the kernel module,
 *      every other symbol can be computing by adding the base
 *      to the output of nm.
*
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int VMX86_RegisterMonitor(int);

#ifdef KERNEL_2_1
EXPORT_SYMBOL(VMX86_RegisterMonitor);
#else
static struct symbol_table vmx86_syms = {
#include <linux/symtab_begin.h>
   X(VMX86_RegisterMonitor),
#include <linux/symtab_end.h>
};
#endif

int
VMX86_RegisterMonitor(int value)
{
   printk("/dev/vmmon: RegisterMonitor(%d) \n",value);
   return 1291;
}


/*
 *----------------------------------------------------------------------
 *
 * init_module --
 *
 *      linux module entry point. Called by /sbin/insmod command
 *
 * Results:
 *      registers a device driver for a major # that depends
 *      on the uid. Add yourself to that list.  List is now in
 *      private/driver-private.c.
 *
 *----------------------------------------------------------------------
 */
int
init_module(void)
{
   int retval;

   /*
    * Call this first, because Warning, Log, etc. try to
    * grab a spin lock.
    */
   HostIF_InitSpinLock();

   if (!LinuxDriverCheckPadding()) {
      return -ENOEXEC;
   }

   CPUID_Init();
   Task_ProbeSysenterMSR();

   /*
    * Initialize LinuxDriverPoll state
    */

   init_waitqueue_head(&linuxState.pollQueue);
   init_timer(&linuxState.pollTimer);
   linuxState.pollTimer.data = 0;
   linuxState.pollTimer.function = LinuxDriverPollTimeout;

   /*
    * Initialize the file_operations structure. Because this code is always
    * compiled as a module, this is fine to do it here and not in a static
    * initializer --hpreg
    */

   memset(&vmuser_fops, 0, sizeof vmuser_fops);
   compat_fop_set_owner(&vmuser_fops);
   compat_fop_set_poll(&vmuser_fops, LinuxDriverPoll);
   vmuser_fops.ioctl = LinuxDriver_Ioctl;
   vmuser_fops.open = LinuxDriver_Open;
   vmuser_fops.release = LinuxDriver_Close;
   vmuser_fops.mmap = LinuxDriverMmap;

#ifdef VMX86_DEVEL
   devel_init_module();
   linuxState.minor = 0;
   retval = register_chrdev(linuxState.major, linuxState.deviceName,
			    &vmuser_fops);
#else
   sprintf(linuxState.deviceName, "vmmon");
   linuxState.major = 10;
   linuxState.minor = 165;
   linuxState.misc.minor = linuxState.minor;
   linuxState.misc.name = linuxState.deviceName;
   linuxState.misc.prev = 0;
   linuxState.misc.next = 0;
   linuxState.misc.fops = &vmuser_fops;

   retval = misc_register(&linuxState.misc);
#endif

   if (retval) {
      Warning("Module %s: error registering with major=%d minor=%d\n",
	      linuxState.deviceName, linuxState.major, linuxState.minor);
      return -ENOENT;
   }
   Log("Module %s: registered with major=%d minor=%d\n",
       linuxState.deviceName, linuxState.major, linuxState.minor);

#ifndef KERNEL_2_1
   retval = register_symtab(&vmx86_syms);
#endif

   /*
    * Snap shot the time stamp counter and the real time so we
    * can later compute an estimate of the cycle time.
    */

   Vmx86_SetStartTime(&linuxState.startTime);

   Log("Module %s: initialized\n", linuxState.deviceName);

   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * cleanup_module --
 *
 *      Called by /sbin/rmmod
 *
 *
 *----------------------------------------------------------------------
 */

void
cleanup_module(void)
{
   int retval;

   /*
    * XXX smp race?
    */
#ifdef VMX86_DEVEL
   retval =  unregister_chrdev(linuxState.major, linuxState.deviceName);
#else
   retval = misc_deregister(&linuxState.misc);
#endif

   if (retval) {
      Warning("Module %s: error unregistering\n", linuxState.deviceName);
   } else {
      Log("Module %s: unloaded\n", linuxState.deviceName);
   }

   compat_del_timer_sync(&linuxState.pollTimer);
}



/*
 *----------------------------------------------------------------------
 *
 * LinuxDriver_Open  --
 *
 *      called on open of /dev/vmmon or /dev/vmx86.$USER. Use count used
 *      to determine eventual deallocation of the module
 *
 * Side effects:
 *     Increment use count used to determine eventual deallocation of
 *     the module
 *
 *----------------------------------------------------------------------
 */
static int
LinuxDriver_Open(struct inode *inode, // IN
                 struct file *filp)   // IN
{
   VMLinux *vmLinux;

   compat_mod_inc_refcount;

   vmLinux = kmalloc(sizeof *vmLinux, GFP_KERNEL);
   if (vmLinux == NULL) {
      compat_mod_dec_refcount;
      return -ENOMEM;
   }
   memset(vmLinux, 0, sizeof *vmLinux);

   sema_init(&vmLinux->lock4Gb, 1);

   filp->private_data = vmLinux;
   LinuxDriverQueue(vmLinux);

#ifdef SUPPORT_LINUXVMWARE
   VMWare_SDumpInit();
   VMWare_SDumpSend("Real men use serial ports! \n");
#endif

#ifdef DO_APM
   apm_register_callback(&LinuxDriverAPMCallback);
#endif

   return 0;
}


/*
 *-----------------------------------------------------------------------------
 *
 * LinuxDriverDestructor4Gb --
 *
 *    Deallocate all directly mappable memory.
 *
 * Results:
 *    None
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

static void
LinuxDriverDestructor4Gb(VMLinux *vmLinux) // IN
{
   unsigned int pg;

   if (!vmLinux->size4Gb) {
      return;
   }
   for (pg = 0; pg < vmLinux->size4Gb; pg++) {
      HostIF_FreePage(vmLinux->pages4Gb[pg]);
   }
   vmLinux->size4Gb = 0;
}


/*
 *----------------------------------------------------------------------
 *
 * LinuxDriver_Close  --
 *
 *      called on close of /dev/vmmon or /dev/vmx86.$USER, most often when the
 *      process exits. Decrement use count, allowing for possible uninstalling
 *      of the module.
 *
 *----------------------------------------------------------------------
 */

#ifdef KERNEL_2_1
static int
#else
static void
#endif
LinuxDriver_Close(struct inode *inode, // IN
                  struct file *filp)   // IN
{
   VMLinux *vmLinux;

   vmLinux = (VMLinux *)filp->private_data;
   ASSERT(vmLinux);

   VMWare_SetVTracer(0);
   LinuxDriverDequeue(vmLinux);
   if (vmLinux->vm != NULL) {
      Vmx86_ReleaseVM(vmLinux->vm);
      vmLinux->vm = NULL;
   }

   /*
    * Destroy all low memory allocations. We are closing the struct file here, so
    * clearly no other process uses it anymore, and we do not need to hold the
    * semaphore --hpreg
    */
   LinuxDriverDestructor4Gb(vmLinux);

   kfree(vmLinux);
   filp->private_data = NULL;
   compat_mod_dec_refcount;
#ifdef KERNEL_2_1
   return 0;
#endif
}


/*
 *----------------------------------------------------------------------
 *
 * LinuxDriverPoll  --
 *
 *      This is used to wake up the VMX when a user call arrives, or
 *      to wake up select() or poll() at the next clock tick.
 *
 *----------------------------------------------------------------------
 */

unsigned int
LinuxDriverPoll(struct file *filp,
		poll_table *wait)
{
   VMLinux *vmLinux = (VMLinux *) filp->private_data;
   unsigned int mask;
   VMDriver *vm = vmLinux->vm;

   if (vm != NULL) {
      /*
       * Check for user call requests.
       */

      if (wait != NULL) {
         poll_wait(filp, &vm->vmhost->callQueue, wait);
      }
      mask = atomic_read(&vm->vmhost->pendingUserCalls) > 0 ? POLLIN : 0;

   } else {
      /*
       * Set up or check the timeout for fast wakeup.
       *
       * Thanks to Petr for this simple and correct implementation:
       *
       * There are three cases of wait == NULL:
       *    another file descriptor is ready in the same poll()
       *    just slept and woke up
       *    nonblocking poll()
       * In all of these, it's okay to return POLLIN.
       *
       * We may schedule a timer unnecessarily if an existing
       * timer fires between poll_wait() and timer_pending().
       *
       * -- edward
       */

      if (wait == NULL) {
	 mask = POLLIN;
      } else {
	 poll_wait(filp, &linuxState.pollQueue, wait);
         if (!timer_pending(&linuxState.pollTimer)) {
	    mod_timer(&linuxState.pollTimer, jiffies + 1);
	 }
	 mask = 0;
      }
   }
   return mask;
}


/*
 *----------------------------------------------------------------------
 *
 * LinuxDriverPollTimeout  --
 *
 *      Wake up a process waiting in poll/select.  This is called from
 *      the timer, and hence processed in the bottom half
 *
 *----------------------------------------------------------------------
 */

void
LinuxDriverPollTimeout(unsigned long clientData)
{
   wake_up(&linuxState.pollQueue);
}


/*
 *----------------------------------------------------------------------
 *
 * LinuxDriverIPIHandler  --
 *
 *      Null IPI handler - for monitor to notice AIO completion
 *
 *----------------------------------------------------------------------
 */
void
LinuxDriverIPIHandler(void *info)
{
   return;
}


/*
 *-----------------------------------------------------------------------------
 *
 * LinuxDriverNoPage --
 *
 *      Callback for returning allocated page for memory mapping
 *
 * Results:
 *
 *      Page or page address on success, NULL or 0 on failure.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#ifdef KERNEL_2_4_0
static struct page *LinuxDriverNoPage(struct vm_area_struct *vma, //IN
				      unsigned long address, 	  //IN
				      int unused)		  //nothing
#else
static unsigned long LinuxDriverNoPage(struct vm_area_struct *vma,//IN
				       unsigned long address,	  //IN
				       int write_access)	  //IN
#endif
{
	VMLinux *vmLinux = (VMLinux *) vma->vm_file->private_data;
	unsigned long pg;
	
	pg = (address - vma->vm_start) >> PAGE_SHIFT;
	if (pg > vmLinux->size4Gb) {
		return 0;
	}
	get_page(virt_to_page(vmLinux->pages4Gb[pg]));
#ifdef KERNEL_2_4_0
	return virt_to_page(vmLinux->pages4Gb[pg]);
#else
	return (unsigned long)vmLinux->pages4Gb[pg];
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * LinuxDriverMmap --
 *
 *      Create mapping for low memory areas.
 *
 * Results:
 *
 *      0       on success,
 *      -EINVAL on invalid arguments or
 *      -ENOMEM on out of memory
 *
 * Side effects:
 *      Pages for mapping are allocated.
 *
 *-----------------------------------------------------------------------------
 */

static int LinuxDriverMmap(struct file *filp, struct vm_area_struct *vma) {
	VMLinux *vmLinux = (VMLinux *) filp->private_data;
	unsigned long size;
	unsigned long i;

	/* Only shared mappings */
	if (!(vma->vm_flags & VM_SHARED)) {
		return -EINVAL;
	}
	if ((vma->vm_end | vma->vm_start) & (PAGE_SIZE - 1)) {
		return -EINVAL;
	}
	size = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;
	if (size < 1) {
		return -EINVAL;
	}
	/* 16 pages looks like a good limit... */
	if (size > VMMON_MAX_LOWMEM_PAGES) {
		return -ENOMEM;
	} 
	/* Sorry. Only one mmap per one open. */
	down(&vmLinux->lock4Gb);
	if (vmLinux->size4Gb) {
		up(&vmLinux->lock4Gb);
		return -EINVAL;
	}
	for (i = 0; i < size; i++) {
		vmLinux->pages4Gb[i] = HostIF_AllocPage();
		if (!vmLinux->pages4Gb[i]) {
			while (i > 0) {
				HostIF_FreePage(vmLinux->pages4Gb[--i]);
			}
			up(&vmLinux->lock4Gb);
			return -ENOMEM;
		}
		memset(vmLinux->pages4Gb[i], 0, PAGE_SIZE);
	}
	vmLinux->size4Gb = size;
	up(&vmLinux->lock4Gb);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 3)
	vma->vm_file = filp;
	filp->f_count++;
#endif
	vma->vm_ops = &vmuser_mops;
	return 0;
}


/*
 *-----------------------------------------------------------------------------
 *
 * LinuxDriver_Ioctl --
 *
 *      Main path for UserRPC
 *
 * Results:
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#include <linux/sched.h>
static int
LinuxDriver_Ioctl(struct inode *inode,
                  struct file *filp,
                  u_int iocmd,
                  unsigned long ioarg)
{
   VMLinux *vmLinux = (VMLinux *) filp->private_data;
   int retval = 0;
   Vcpuid vcpuid;

   VMWare_SetVTracer(VTrace_Set);

   switch (iocmd) {
   case IOCTLCMD_VERSION:
      retval = VMMON_VERSION;
      break;

   case IOCTLCMD_CREATE_VM: {
      WslimitsInfo wslimitsInfo;
      if (vmLinux->vm != NULL) {
	 retval = -EINVAL;
	 break;
      }

      memset(&wslimitsInfo, 0, sizeof wslimitsInfo);
      if ((char *) ioarg != NULL) {
          if (HostIF_CopyFromUser(&wslimitsInfo, (char *) ioarg,
                                  sizeof wslimitsInfo)) {
             Log("CopyFromUser failed for WslimitsInfo struct \n");
          }
      }
      vmLinux->vm = Vmx86_CreateVM((void *)filp, current->pid, 
				   &wslimitsInfo);
      if (vmLinux->vm == NULL) {
	 retval = -ENOMEM;
	 break;
      }
      retval = vmLinux->vm->id;
      break;
   }
   case IOCTLCMD_BIND_VM:
      if (vmLinux->vm != NULL) {
	 retval = -EINVAL;
	 break;
      }
      vmLinux->vm = Vmx86_BindVM((int) ioarg);
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      break;

   case IOCTLCMD_RELEASE_VM: {
      VMDriver *vm;
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      vm = vmLinux->vm;
      vmLinux->vm = NULL;
      Vmx86_ReleaseVM(vm);
      break;
   }

   case IOCTLCMD_INIT_VM: {
      InitBlock initParams;
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CopyFromUser(&initParams, (char*)ioarg,
				   sizeof initParams);
      if (retval != 0) {
	 break;
      }
      if (Vmx86_InitVM(vmLinux->vm, &initParams)) {
	 retval = -EINVAL;
	 break;
      }
      break;
   }

   case IOCTLCMD_LATE_INIT_VM:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      if (Vmx86_LateInitVM(vmLinux->vm)) {
	 retval = -EINVAL;
	 break;
      }
      break;

   case IOCTLCMD_RUN_VM:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      vcpuid = (Vcpuid) ioarg;
      if (vcpuid >= vmLinux->vm->numVCPUs) {
         retval = -EINVAL;
         break;
      }
#ifdef __SMP__
      unlock_kernel();
#endif
      retval = Vmx86_RunVM(vmLinux->vm, vcpuid);
      ASSERT(retval >= 0 || retval == USERCALL_RESTART);
#ifdef __SMP__
      lock_kernel();
#endif
      break;

   case IOCTLCMD_SET_UID:
#ifdef VMX86_DEVEL
      devel_suid();
#else
      retval = -EPERM;
#endif
      break;

   case IOCTLCMD_LOCK_PAGE:
   case IOCTLCMD_UNLOCK_PAGE:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
   case IOCTLCMD_LOOK_UP_MPN: {
      void *addr = (char *)ioarg;
      ASSERT(sizeof (int) == sizeof (MPN));
      retval = iocmd == IOCTLCMD_LOOK_UP_MPN ?
		  (int) HostIF_LookupUserMPN(vmLinux->vm, addr) :
	       iocmd == IOCTLCMD_LOCK_PAGE ?
		  (int) Vmx86_LockPage(vmLinux->vm, addr, TRUE) :
		  Vmx86_UnlockPage(vmLinux->vm, addr, TRUE);
      /*
       * Since ioctl() can't return negative values,
       * encode error code as a positive number.
       */
      if (!PAGE_LOCK_SUCCESS(retval)) {
	 ASSERT(1 << (sizeof retval * 8 - PAGE_SHIFT) < PAGE_LOCK_ERROR_BIT);
	 retval = PAGE_LOCK_ERROR_BIT | -retval;
      }
      break;
   }

   case IOCTLCMD_GET_NUM_VMS: {
      retval = Vmx86_GetNumVMs();
      break;
   }

   case IOCTLCMD_GET_TOTAL_MEM_USAGE: {
      retval = Vmx86_GetTotalMemUsage();
      break;
   }

   case IOCTLCMD_SET_HARD_LIMIT: {
      int32 limit;
      retval = HostIF_CopyFromUser(&limit, (void *)ioarg, sizeof limit);
      if (retval != 0) {
	 break;
      }
      if (!Vmx86_SetLockedPagesLimit(limit)) {
         retval = -EINVAL;
      }
      break;
   }

   case IOCTLCMD_GET_HARD_LIMIT: {
      retval = Vmx86_GetLockedPagesLimit();
      break;
   }

   case IOCTLCMD_SET_MEM_INFO: {
      VMSetMemInfoArgs args;

      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CopyFromUser(&args, (void *)ioarg, sizeof args);
      if (retval != 0) {
	 break;
      }
      Vmx86_SetMemInfo(vmLinux->vm, &args);
      break;
   }

   case IOCTLCMD_GET_MEM_INFO: {
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      if (!Vmx86_GetMemInfoCopy(vmLinux->vm, (VMGetMemInfoArgs *) ioarg)) {
         retval = -EINVAL;
      }
      break;
   }

   case IOCTLCMD_GET_STATS: {
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CopyToUser((void *)ioarg, &vmLinux->vm->stats,
                                 sizeof vmLinux->vm->stats);
      if (retval != 0) {
	 break;
      }
      break;
   }

   case IOCTLCMD_SET_STATS: {
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CopyFromUser(&vmLinux->vm->stats, (void *)ioarg,
				   sizeof vmLinux->vm->stats);
      break;
   }

   case IOCTLCMD_PAE_ENABLED:
      retval = Vmx86_PAEEnabled();
      break;

   case IOCTLCMD_IS_MP_SAFE: {
#ifdef __SMP__
      retval = TRUE;
#else
      retval = FALSE;
#endif
      break;
   }

   case IOCTLCMD_APIC_BASE: {
      VMAPICInfo info; 
      Bool setVMPtr; 
      Bool probe;
#ifdef USE_PERFCTRS_HOSTED
      Bool forceEnable;
#endif

      retval = HostIF_CopyFromUser(&info, (VMAPICInfo *)ioarg,
                                   sizeof info);
      if (retval != 0) {
         break; 
      }
      setVMPtr = ((info.flags & APIC_FLAG_DISABLE_NMI) != 0);
      probe = ((info.flags & APIC_FLAG_PROBE) != 0);
#ifdef USE_PERFCTRS_HOSTED
      forceEnable = ((info.flags & APIC_FLAG_FORCE_ENABLE) != 0);
#endif
      
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 20)
      // Kernel uses NMIs for deadlock detection -
      //  set APIC VMptr so that NMIs get disabled in the monitor
      setVMPtr = TRUE;
#endif
      info.apicBase = HostIF_APICBase(vmLinux->vm, setVMPtr, probe);

#ifdef USE_PERFCTRS_HOSTED
      if ((info.apicBase == 0) && forceEnable) {
         info.apicBase = HostIF_APICEnable(vmLinux->vm);
      }
#endif

      retval = HostIF_CopyToUser((VMAPICInfo *)ioarg, &info, sizeof info);
      break;
   }

   case IOCTLCMD_IOAPIC_BASE: {
      MA ma; 

      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      ma = HostIF_IOAPICBase(vmLinux->vm);
      retval = HostIF_CopyToUser((MA *)ioarg, &ma, sizeof ma);
      break;
   }

   case IOCTLCMD_CHECK_MEMORY:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CheckMemory(vmLinux->vm);
      break;

#ifdef SUPPORT_PASSTHROUGH
   /*
    * Support for passthru was removed in 01/2002. The ioctl
    * commands are kept for backwards compatibility of driver.
    * --Ganesh 01/2001.
    */
   case IOCTLCMD_REGISTER_PASSTHROUGH_IRQ: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_REGISTER_PASSTHROUGH_IO: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_FREE_PASSTHROUGH_IRQ: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_FREE_PASSTHROUGH_IO: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_START_PASSTHROUGH: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_STOP_PASSTHROUGH: {
      retval = -EINVAL;
      break;
   }

   case IOCTLCMD_QUERY_PASSTHROUGH: {
      retval = -EINVAL;
      break;
   }
#endif // ifdef SUPPORT_PASSTHROUGH

   case IOCTLCMD_ALLOW_CORE_DUMP:
      if (current->euid == current->uid &&
	  current->fsuid == current->uid &&
          current->egid == current->gid &&
	  current->fsgid == current->gid) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 7)
	 current->mm->dumpable = 1;
#else
	 current->dumpable = 1;
#endif
         retval = 0;
      } else {
         retval = -EPERM;
      }
      break;

   case IOCTLCMD_BROADCAST_IPI:
      retval = compat_smp_call_function(LinuxDriverIPIHandler, NULL, FALSE,
                               FALSE);
      break;

#ifdef USE_PERFCTRS_HOSTED
   case IOCTLCMD_REGISTER_PERFCTR: {
      PerfCtrRegisterArgs args;

      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = HostIF_CopyFromUser(&args, (void *)ioarg, sizeof args);
      if (retval != 0) {
	 break;
      }
      retval = PerfCtr_Register(vmLinux->vm, &args);
      break;
   }
   case IOCTLCMD_START_PERFCTR:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = PerfCtr_StartCounters(vmLinux->vm);
      break;

   case IOCTLCMD_STOP_PERFCTR:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = PerfCtr_StopCounters(vmLinux->vm);
      break;

   case IOCTLCMD_RELEASE_PERFCTR:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      retval = PerfCtr_Release(vmLinux->vm, TRUE);
      break;
#endif // ifdef USE_PERFCTRS_HOSTED

   case IOCTLCMD_ALLOC_LOW_PAGES:
      retval = -EINVAL;
      break;

   case IOCTLCMD_FREE_LOW_PAGES:
      retval = -EINVAL;
      break;

   case IOCTLCMD_GET_KHZ_ESTIMATE:
      retval = Vmx86_GetkHzEstimate(&linuxState.startTime);
      break;

   case IOCTLCMD_GET_SYSTEM_CLOCK_RATE:
      retval = HZ;
      break;

   case IOCTLCMD_ACK_USER_CALL:
      vcpuid = (Vcpuid) ioarg;
      if (vcpuid >= vmLinux->vm->numVCPUs) {
         retval = -EINVAL;
         break;
      }
      Vmx86_AckUserCall(vmLinux->vm, vcpuid);
      break;

   case IOCTLCMD_COMPLETE_USER_CALL:
      if (vmLinux->vm == NULL) {
	 retval = -EINVAL;
	 break;
      }
      vcpuid = (Vcpuid) ioarg;
      if (vcpuid >= vmLinux->vm->numVCPUs) {
         retval = -EINVAL;
         break;
      }
      Vmx86_CompleteUserCall(vmLinux->vm, vcpuid);
      break;

   default:
      Warning("Unknown ioctl %d\n", iocmd);
      retval = -EINVAL;
   }

   VMWare_SetVTracer(0);
   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * vLog --
 *
 *     Log() from the kernel module logged to fd
 *
 *
 *----------------------------------------------------------------------
 */

static INLINE void
vLog(int fd) // IN
{
   struct file *file;

   ASSERT(strlen(linuxState.buf));
   ASSERT(strlen(linuxState.buf) < LINUXLOG_BUFFER_SIZE);

   file = vmware_fget(fd);
   if (file) {
#ifdef KERNEL_2_1
      mm_segment_t old_fs = get_fs();

      set_fs(get_ds());
      file->f_op->write(file, (char*)linuxState.buf, strlen(linuxState.buf),
                        &file->f_pos);
#else
      int32 old_fs = get_fs();

      set_fs(get_ds());
      file->f_op->write(file->f_inode, file, (char*)linuxState.buf,
                        strlen(linuxState.buf));
#endif
      set_fs(old_fs);
      compat_fput(file);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * vWarning --
 *
 *      Warning() get here. Adds a \r to compensate for the
 *      uncooked mode of the terminal.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
vWarning(VMDriver *vm) // IN
{
   struct file *file;

   ASSERT(strlen(linuxState.buf));
   ASSERT(strlen(linuxState.buf) < LINUXLOG_BUFFER_SIZE);

   if (vm && (file = vmware_fget(1))) {
#ifdef KERNEL_2_1
      mm_segment_t old_fs = get_fs();

      set_fs(get_ds());
      file->f_op->write(file, (char*)linuxState.buf, strlen(linuxState.buf),
                        &file->f_pos);
#else
      int32 old_fs = get_fs();

      set_fs(get_ds());
      file->f_op->write(file->f_inode, file, (char*)linuxState.buf,
                        strlen(linuxState.buf));
#endif
      set_fs(old_fs);
      compat_fput(file);

      printk("/dev/vmmon: (pid=%d) %s", current->pid, linuxState.buf);

      return;
   }

   /* Use the kernel log */
   printk(KERN_WARNING "/dev/vmmon: %s", linuxState.buf);
}


/*
 *----------------------------------------------------------------------
 *
 * Warning --
 *
 *      Warning messages from kernel module: logged to stdout and the log file
 *
 *----------------------------------------------------------------------
 */
void
Warning(char *fmt,...)
{
   VMDriver *vm;
   va_list args;

   vm = Vmx86_GetVMforProcess((uintptr_t)(current->pid));

   va_start(args, fmt);
   vsprintf(linuxState.buf, fmt, args);
   va_end(args);

   if (vm) {
      vLog(vm->logFD);
   }
   vWarning(vm);
}

/*
 *----------------------------------------------------------------------
 *
 * Log --
 *
 *      Log messages from kernel module: logged to log file only
 *
 *----------------------------------------------------------------------
 */
void
Log(char *fmt,...)
{
   VMDriver *vm;
   va_list args;


   vm = Vmx86_GetVMforProcess((uintptr_t)(current->pid));

   va_start(args, fmt);
   vsprintf(linuxState.buf, fmt, args);
   va_end(args);

   if (vm) {
      vLog(vm->logFD);
   } else {
      /* Use the kernel log with at least a KERN_DEBUG level so that it doesn't
         garbage the screen at (re)boot time on RedHat 6.0 --hpreg */
      printk(KERN_DEBUG "/dev/vmmon: %s", linuxState.buf);
   }
}


/*
 *----------------------------------------------------------------------
 *
 * Panic --
 *
 *      ASSERTION failures and Panics from kernel module get here.
 *      Message is logged to stdout and the log file
 *
 *
 * Side effects:
 *      Never returns
 *
 *----------------------------------------------------------------------
 */
void
Panic(char *fmt, ...)
{
   asmlinkage int sys_exit(int exit_code);
   VMDriver *vm = Vmx86_GetVMforProcess((uintptr_t)(current->pid));
   va_list args;

   va_start(args, fmt);
   vsprintf(linuxState.buf, fmt, args);
   va_end(args);

   /*
    * XXX
    * XXX We cannot exit() the process since we are not running it
    * XXX
    */
#ifdef KERNEL_2_1
   if (in_interrupt()) {
      printk("/dev/vmmon: Panic in interrupt (no intrcount)\n");
      panic("Assertion failure in interrupt handling in VMX86\n");
   }

#else
   if (intr_count) {
      printk("/dev/vmmon: Panic intrcount=%ld\n",intr_count);
      panic("Assertion failure in interrupt handling in VMX86\n");
   }
#endif

   if (vm) {
      vLog(vm->logFD);
   }
   vWarning(vm);
   sprintf(linuxState.buf,"VMX86 driver panic. pid=%d\n\r",current->pid);
   if (vm) {
      vLog(vm->logFD);
   }
   vWarning(vm);

   _exit(1);
   for (;;);   /* To supress warning. */
}



/*
 *----------------------------------------------------------------------
 *
 * LinuxDriverQueue --
 *
 *      add the vmLinux to the global queue
 *
 * Results:
 *
 *      void
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
static void
LinuxDriverQueue(VMLinux *vmLinux)
{
   /*
    * insert in global vm queue
    */

   HostIF_GlobalVMLock(12);

   vmLinux->next = linuxState.head;
   linuxState.head = vmLinux;


   HostIF_GlobalVMUnLock(12);

}


/*
 *----------------------------------------------------------------------
 *
 * LinuxDriveDequeue --
 *
 *      remove from active list
 *
 * Results:
 *
 *      void
 * Side effects:
 *      printk if it is not in the list (error condition)
 *
 *----------------------------------------------------------------------
 */
static void
LinuxDriverDequeue(VMLinux *vmLinux)
{
   VMLinux **p;

   HostIF_GlobalVMLock(13);
   for (p = &linuxState.head; *p != vmLinux; p = &(*p)->next) {
      ASSERT(*p != NULL);
   }
   *p = vmLinux->next;
   vmLinux->next = NULL;
   HostIF_GlobalVMUnLock(13);
}


/*
 *----------------------------------------------------------------------
 *
 * CheckPadding --
 *
 *      check for expected padding --
 *      this check currently fails on the egcs compiler
 *
 * Results:
 *
 *      TRUE if the check succeeds -- module will be loaded
 *
 *
 *
 * Side effects:
 *      output to kernel log on error
 *
 *----------------------------------------------------------------------
 */
static Bool
LinuxDriverCheckPadding(void)
{
   DTRWords dtr;
   uint16 *x;

   memset(&dtr, 0, sizeof dtr);
   dtr.dtr.limit = 0x1111;
   dtr.dtr.offset = 0x22223333;

   x = (uint16*)&dtr;

   if (x[0] == 0x1111 && x[1] == 0x3333 && x[2] == 0x2222) {
   } else {
      Warning("DTR padding\n");
      goto error;
   }

   if (   sizeof (VMCrossPage) != MODULECALL_CROSSPAGE_SIZE
       || offsetof(VMCrossPage, crosspageLinearAddr)
             != MODULECALL_CROSSPAGE_LAOFFSET) {
      Warning("cross page 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof (VMCrossPage), MODULECALL_CROSSPAGE_SIZE,
              offsetof(VMCrossPage, crosspageLinearAddr),
              MODULECALL_CROSSPAGE_LAOFFSET);
      goto error;
   }

   return TRUE;

error:
   printk("/dev/vmmon: Cannot load module. Use standard gcc compiler\n");
   return FALSE;
}


#ifdef DO_APM
static int LinuxDriverAPMCallback(apm_event_t event)
{
   switch (event) {
   case APM_SYS_SUSPEND:
   case APM_USER_SUSPEND:
      if (LinuxDriverAPMstate == APM_STATE_READY) {
	 // call up to user to suspend VMs
      }
      break;
   case APM_NORMAL_RESUME:
   case APM_CRITICAL_RESUME:
      if (LinuxDriverAPMstate == APM_STATE_SUSPEND) {
	 // call up to user to resume VMs
      }
   }
   return 0;
}
#endif
