/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

#ifndef _VNET_H
#define _VNET_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_MONITOR
#include "includeCheck.h"
#include "vm_basic_types.h"
#include "vm_atomic.h"

#define SIOCSKEEP          0x89F0  // not used
#define SIOCGKEEP          0x89F1  // not used
#define SIOCSLADRF         0x89F2
#define SIOCPORT           0x89F3
#define SIOCBRIDGE         0x89F4
#define SIOCNETIF          0x89F5

/*
 *  16 private IOCTLs available. However SIOCDEVPRIVATE == 0x89F0
 *  so we need to start at 6 to aviod above IOCTLs.
 */

#define SIOCSETMACADDR     0x89F6
#define SIOCSSWITCHMAP     0x89F7
#define SIOCSETNOTIFY      0x89F8
#define SIOCUNSETNOTIFY    0x89F9
#define SIOCSETCLUSTERSIZE 0x89FA
#define SIOCSETNOTIFY2     0x89FB
#define SIOCGETAPIVERSION  0x89FC

/*
 * VMnet driver version.
 *
 * Increment major version when you make an incompatible change.
 * Compatibility goes both ways (old driver with new executable
 * as well as new driver with old executable).
 */

#ifdef linux
#define VNET_API_VERSION		(2 << 16 | 0)
#else
#define VNET_API_VERSION                (3 << 16 | 0)
#endif
#define VNET_API_VERSION_MAJOR(v)	((uint32) (v) >> 16)
#define VNET_API_VERSION_MINOR(v)	((uint16) (v))

/* version 1 structure */

typedef struct VNet_SetMacAddrIOCTL {
   int             version;
   unsigned char   addr[6];
   unsigned        flags;
} VNet_SetMacAddrIOCTL;

typedef struct VNet_Notify {
   int              version;
   Atomic_uint32   *actPtr;
   uint32 volatile *pollPtr;
   uint32           actMask;
   uint32           pollMask;
} VNet_Notify;

#define VNET_SETMACADDRF_UNIQUE      0x01
/*
 * The latest 802.3 standard sort of says that the length field ought to
 * be less than 1536 (for VLAN tagging support). I am choosing 1532
 * as our max VNET_MTU size, as I'd rather keep it a multiple of 4 and
 * VLAN tagging uses only upto 1518 bytes.
 */
#define VNET_MTU                     1532


#define VNET_BUF_TOO_SMALL           -1

/*
 *  vlan switch stuff
 */


#define VNET_MAX_VLANS     255

struct VNetSwitchMap {
   int  trunk;
   int  vlan;
   int  connect;
   int  vnet;
};

#endif
